@@ -320,11 +320,19 @@ static struct tl_root *parse_edl(bstr str, struct mp_log *log)
         }
     }
     mp_assert(root->num_pars);
+#ifdef FUZZING_BUILD_MODE_UNSAFE_FOR_PRODUCTION
+    if (root->num_pars > 3)
+        goto error;
+#endif
     for (int n = 0; n < root->num_pars; n++) {
         if (root->pars[n]->num_parts < 1) {
             mp_err(log, "EDL specifies no segments.'\n");
             goto error;
         }
+#ifdef FUZZING_BUILD_MODE_UNSAFE_FOR_PRODUCTION
+        if (root->pars[n]->num_parts > 3)
+            goto error;
+#endif
     }
     return root;
 error:
@@ -344,6 +352,7 @@ static struct demuxer *open_source(struct timeline *root,
     struct demuxer_params params = {
         .init_fragment = tl->init_fragment,
         .stream_flags = root->stream_origin,
+        .depth = root->demuxer->depth + 1,
     };
     struct demuxer *d = demux_open_url(filename, &params, root->cancel,
                                        root->global);
@@ -435,6 +444,7 @@ static struct timeline_par *build_timeline(struct timeline *root,
         struct demuxer_params params = {
             .init_fragment = tl->init_fragment,
             .stream_flags = root->stream_origin,
+            .depth = root->demuxer->depth + 1,
         };
         tl->track_layout = demux_open_url("memory://", &params, root->cancel,
                                           root->global);
