@@ -27,7 +27,7 @@ void terminal_get_size2(int *rows, int *cols, int *px_width, int *px_height)
 {
 }
 
-PRINTF_ATTRIBUTE(2, 0)
+MP_PRINTF_ATTRIBUTE(2, 0)
 int mp_console_vfprintf(void *wstream, const char *format, va_list args)
 {
     return 0;
